from time import sleep
from mrq.task import Task
from mrq.context import log, connections
import urllib2


class MongoTimeout(Task):

  def run(self, params):

    res = connections.mongodb_jobs.eval("""
      function() {
        var a;
        for (i=0;i<10000000;i++) { 
          for (y=0;y<10000000;y++) {
            a = Math.max(y);
          }
        }
        return a;
      }
    """)

    return res




#!/usr/bin/env python
# from gevent import monkey
# monkey.patch_all()

import sys
import os
import argparse
import json

sys.path.append(os.getcwd())

from mrq import config, queue, utils
from mrq.context import set_current_config


parser = argparse.ArgumentParser(description='Runs a task')


parser.add_argument('--quiet', action='store_true', default=False, help='No logging')
parser.add_argument('--async', action='store_true', default=False, help='Queue the task instead of running it right away')
parser.add_argument('--queue', action='store', default="default", help='Queue where to put the task when async')
parser.add_argument('taskpath', action='store', help='Task to run')
parser.add_argument('taskargs', action='store', default='{}', nargs='*', help='JSON-encoded arguments, or "key value" pairs')

args = parser.parse_args()

if len(args.taskargs) == 1:
  params = json.loads(args.taskargs[0])
else:
  params = {}

  # mrq-run taskpath a 1 b 2 => {"a": "1", "b": "2"}
  for group in utils.group_iter(args.taskargs, n=2):
    if len(group) != 2:
      print "Number of arguments wasn't even"
      sys.exit(1)
    params[group[0]] = group[1]


def main():

  cfg = config.get_config(sources=("file", "env"))
  cfg["quiet"] = args.quiet
  cfg["is_cli"] = True
  set_current_config(cfg)

  ret = queue.send_task(args.taskpath, params, sync=not args.async, queue=args.queue)
  if args.async:
    print ret
  else:
    print json.dumps(ret)
if __name__ == "__main__":
  main()



from .utils import load_class_by_path, group_iter
from .context import connections, get_current_config


class Queue(object):

  def __init__(self, queue_id):
    if isinstance(queue_id, Queue):
      self.id = queue_id.id  # TODO use __new__?
    else:
      self.id = queue_id

  @property
  def redis_key(self):
    return "%s:q:%s" % (get_current_config()["redis_prefix"], self.id)

  def enqueue_job_ids(self, job_ids):
    connections.redis.rpush(self.redis_key, *job_ids)

  def size(self):
    return connections.redis.llen(self.redis_key)

  def list_job_ids(self, skip=0, limit=20):
    return connections.redis.lrange(self.redis_key, skip, skip + limit - 1)

  @classmethod
  def all(cls):
    # TODO MongoDB distinct?
    prefix = get_current_config()["redis_prefix"]
    queues = []
    for key in connections.redis.keys():
      if key.startswith(prefix):
        queues.append(Queue(key[len(prefix) + 3:]))

    return queues


def send_task(path, params, **kwargs):
  return send_tasks(path, [params], **kwargs)[0]


def send_tasks(path, params_list, queue=None, sync=False, batch_size=1000):

  if len(params_list) == 0:
    return []

  if sync:
    task_class = load_class_by_path(path)
    return [task_class().run(params) for params in params_list]

  if queue is None:
    task_def = get_current_config().get("tasks", {}).get(path) or {}
    queue = task_def.get("queue", "default")

  all_ids = []

  for params_group in group_iter(params_list, n=batch_size):

    collection = connections.mongodb_jobs.mrq_jobs

    job_ids = collection.insert([{
      "path": path,
      "params": params,
      "queue": queue,
      "status": "queued"
    } for params in params_group], w=1)

    # Between these 2 calls, a task can be inserted in MongoDB but not queued in Redis.
    # This is the same as dequeueing a task from Redis and being stopped before updating the "started"
    # flag in MongoDB.

    Queue(queue).enqueue_job_ids([str(x) for x in job_ids])

    all_ids += job_ids

  return all_ids


    # # Cf code in send_task
    # if kwargs.get("uniquestarted") or kwargs.get("uniquequeued"):
    #   descriptions = {json.dumps([task, params, {}]):i for i, params in enumerate(params_list)}
    #   cancelled_indexes = []

    #   if kwargs.get("uniquestarted"):
    #     del kwargs["uniquestarted"]
    #     started_queue = Queue("started", connection=connection)

    #     for job in started_queue.jobs:
    #       if job.description in descriptions:
    #         cancelled_indexes.append(descriptions.get(job.description))

    #   if kwargs.get("uniquequeued"):
    #     params_list = uniqify_params_list(params_list)
    #     del kwargs["uniquequeued"]
    #     started_queue = Queue(task, connection=connection)

    #     for job in started_queue.jobs:
    #       if job.description in descriptions and job.status != "finished":
    #         cancelled_indexes.append(descriptions.get(job.description))

    #   # http://stackoverflow.com/questions/11303225/how-to-remove-multiple-indexes-from-a-list-at-the-same-time
    #   for i in sorted(set(cancelled_indexes), reverse=True):
    #     params_list.pop(i)



#!/usr/bin/env python
from gevent import monkey
monkey.patch_all()

import sys
import os
import tempfile
import signal
import subprocess32 as subprocess
import psutil

sys.path.append(os.getcwd())

from mrq import config
from mrq.utils import load_class_by_path


def main():

  cfg = config.get_config()

  # If we are launching with a --processes option and without the SUPERVISOR_ENABLED env
  # then we should just call supervisord.
  if cfg["processes"] > 0 and not os.environ.get("SUPERVISOR_ENABLED"):

    # We wouldn't need to do all that if supervisord supported environment variables in all its config fields!
    with open(cfg["supervisord_template"], "r") as f:
      conf = f.read()

    fh, path = tempfile.mkstemp(prefix="mrqsupervisordconfig")
    f = os.fdopen(fh, "w")

    # We basically relaunch ourselves, but the config will contain the MRQ_SUPERVISORD_ISWORKER env.
    conf = conf.replace("{{ SUPERVISORD_COMMAND }}", " ".join(sys.argv))
    conf = conf.replace("{{ SUPERVISORD_PROCESSES }}", str(cfg["processes"]))

    f.write(conf)
    f.close()

    try:

      # start_new_session=True avoids sending the current process' signals to the child.
      process = subprocess.Popen(["supervisord", "-c", path], start_new_session=True)

      def sigint_handler(signum, frame):

        # At this point we need to send SIGINT to all workers. Unfortunately supervisord
        # doesn't support this, so we have to find all the children pids and send them the
        # signal ourselves :-/
        # https://github.com/Supervisor/supervisor/issues/179
        #
        psutil_process = psutil.Process(process.pid)
        worker_processes = psutil_process.get_children(recursive=False)

        if len(worker_processes) == 0:
          return process.send_signal(signal.SIGTERM)

        for child_process in worker_processes:
          child_process.send_signal(signal.SIGINT)

        # Second time sigint is used, we should terminate supervisord itself which
        # will send SIGTERM to all the processes anyway.
        signal.signal(signal.SIGINT, sigterm_handler)

        # Wait for all the childs to finish
        for child_process in worker_processes:
          child_process.wait()

        # Then stop supervisord itself.
        process.send_signal(signal.SIGTERM)

      def sigterm_handler(signum, frame):
        process.send_signal(signal.SIGTERM)

      signal.signal(signal.SIGINT, sigint_handler)
      signal.signal(signal.SIGTERM, sigterm_handler)

      process.wait()

    finally:
      os.remove(path)

  # If not, start the actual worker
  else:

    worker_class = load_class_by_path(cfg["worker_class"])

    w = worker_class(cfg)

    exitcode = w.work_loop()

    sys.exit(exitcode)

if __name__ == "__main__":
  main()



import argparse
import os
import sys


def get_config(sources=("file", "env", "args"), env_prefix="MRQ_", defaults=None):

  parser = argparse.ArgumentParser(description='Starts an RQ worker')

  parser.add_argument('--max_jobs', default=0, type=int, action='store',
                      help='Gevent: max number of jobs to do before quitting. Temp workaround for memory leaks')

  parser.add_argument('--gevent', '-g', default=1, type=int, action='store',
                      help='Gevent: max number of greenlets')

  parser.add_argument('--processes', '-p', default=0, type=int, action='store',
                      help='Number of processes to launch with supervisord')

  default_template = os.path.abspath(os.path.join(os.path.dirname(__file__), "supervisord_templates/default.conf"))
  parser.add_argument('--supervisord_template', default=default_template, action='store',
                      help='Path of supervisord template to use')

  parser.add_argument('--mongodebug', action='store_true', default=False,
                      help='Print all Mongo requests')

  parser.add_argument('--objgraph', action='store_true', default=False,
                      help='Start objgraph to debug memory after each task')

  parser.add_argument('--trace_greenlets', action='store_true', default=False,
                      help='Collect stats about each greenlet execution time and switches.')

  parser.add_argument('--profile', action='store_true', default=False,
                      help='Run profiling on the whole worker')

  parser.add_argument('--mongodb_jobs', action='store', default="mongodb://127.0.0.1:27017/mrq",
                      help='MongoDB URI for the jobs database')

  parser.add_argument('--mongodb_logs', action='store', default="mongodb://127.0.0.1:27017/mrq",
                      help='MongoDB URI for the logs database')

  parser.add_argument('--mongodb_logs_size', action='store', default=16 * 1024 * 1024, type=int,
                      help='If provided, sets the log collection to capped to that amount of bytes')

  parser.add_argument('--redis', action='store', default="redis://127.0.0.1:6379",
                      help='Redis URI')

  parser.add_argument('--redis_prefix', action='store', default="mrq",
                      help='Redis key prefix')

  parser.add_argument('--name', default=None, action='store',
                      help='Specify a different name')

  parser.add_argument('--quiet', default=False, action='store_true',
                      help='Don\'t output task logs')

  parser.add_argument('--scheduler', default=False, action='store_true',
                      help='Run the scheduler')

  parser.add_argument('--scheduler_interval', default=60, action='store', type=int,
                      help='Seconds between scheduler checks')

  parser.add_argument('--report_interval', default=10, action='store', type=int,
                      help='Seconds between worker reports to MongoDB')

  parser.add_argument('--config', '-c', default=None, action="store",
                      help='Path of a config file')

  parser.add_argument('--admin_port', default=0, action="store", type=int,
                      help='Start an admin server on this port, if provided. Incompatible with --processes')

  parser.add_argument('--worker_class', default="mrq.worker.Worker", action="store",
                      help='Path to a custom worker class')

  parser.add_argument('queues', nargs='*', default=["default"],
                      help='The queues to listen on (default: \'default\')')

  if "args" in sources:
    from_args = parser.parse_args()
  else:
    from_args = parser.parse_args([])

  # Get defaults
  merged_config = from_args.__dict__
  if defaults is not None:
    merged_config.update(defaults)

  # If a mrq-config.py file is in the current directory, use it!
  default_config_file = os.path.join(os.getcwd(), "mrq-config.py")

  if merged_config["config"] is None and os.path.isfile(default_config_file):
    # print "Using config file at %s" % default_config_file
    merged_config["config"] = default_config_file

  config_module = None
  if "file" in sources and merged_config["config"]:
    sys.path.append(os.path.dirname(merged_config["config"]))
    config_module = __import__(os.path.basename(merged_config["config"].replace(".py", "")))
    sys.path.pop(-1)
    merged_config.update({k.lower(): v for k, v in config_module.__dict__.iteritems() if k[0].isupper()})

  # Keys that can't be passed from the command line
  merged_config["tasks"] = {}
  merged_config["scheduled_tasks"] = {}

  for part in sources:
    for name, arg_value in merged_config.iteritems():

      value = None
      if part == "env":
        value = os.environ.get(env_prefix + name.upper())
      elif part == "args":
        value = arg_value
      elif part == "file":
        value = getattr(config_module, name.upper(), None)

      if value is not None:
        merged_config[name] = value

  return merged_config




from collections import defaultdict
import datetime
import pymongo


class LogHandler(object):
  """ Job/Worker-aware log handler.

      We used the standard logging module before but it suffers from memory leaks
      when creating lots of logger objects.
  """

  def __init__(self, collection=None, quiet=False):
    self.reset()
    self.set_collection(collection)
    self.quiet = quiet

    # Import here to avoid import loop
    from .context import get_current_job
    self.get_current_job = get_current_job

  def get_logger(self, worker=None, job=None):
    return LoggerInterface(self, worker=worker, job=job)

  def set_collection(self, collection=None):
    self.collection = collection

  def reset(self):
    self.buffer = {
      "workers": defaultdict(list),
      "jobs": defaultdict(list)
    }

  def log(self, level, *args, **kwargs):

    worker = kwargs.get("worker")
    job = kwargs.get("job")

    formatted = "%s [%s] %s" % (datetime.datetime.utcnow(), level.upper(), " ".join([unicode(x) for x in args]))

    if not self.quiet:
      print formatted

    if worker is not None:
      self.buffer["workers"][worker].append(formatted)
    elif job is not None:
      if job == "current":
        job_object = self.get_current_job()
        if job_object:
          self.buffer["jobs"][job_object.id].append(formatted)
      else:
        self.buffer["jobs"][job].append(formatted)

  def flush(self, w=0):

    # We may log some stuff before we are even connected to Mongo!
    if self.collection is None:
      return

    inserts = [{
      "worker": k,
      "logs": "\n".join(v) + "\n"
    } for k, v in self.buffer["workers"].iteritems()] + [{
      "job": k,
      "logs": "\n".join(v) + "\n"
    } for k, v in self.buffer["jobs"].iteritems()]

    if len(inserts) == 0:
      return

    self.reset()

    try:
      self.collection.insert(inserts, w=w)
    except pymongo.errors.AutoReconnect:
      from mrq.context import get_current_worker
      self.log("debug", "Log insert failed.", worker=get_current_worker())


class LoggerInterface(object):
  """ This object acts as a logger from python's logging module. """

  def __init__(self, handler, **kwargs):
    self._handler = handler
    self.kwargs = kwargs
    self.quiet = False

  @property
  def handler(self):
    if self._handler:
      return self._handler

    # Import here to avoid import loop
    from .context import get_current_worker
    worker = get_current_worker()

    if worker:
      return worker.log_handler
    else:
      self._handler = LogHandler(quiet=self.quiet)
      return self._handler

  def info(self, *args):
    self.handler.log("info", *args, **self.kwargs)

  def warning(self, *args):
    self.handler.log("warning", *args, **self.kwargs)

  def error(self, *args):
    self.handler.log("error", *args, **self.kwargs)

  def debug(self, *args):
    self.handler.log("debug", *args, **self.kwargs)




import datetime
from bson import ObjectId
import pymongo
import time
from .exceptions import RetryInterrupt, CancelInterrupt
from .utils import load_class_by_path
from .queue import Queue
from .context import get_current_worker, log, connections, get_current_config
import gevent


class Job(object):

  # Seconds the job can last before timeouting
  timeout = 300

  # Seconds the results are kept in MongoDB
  result_ttl = 7 * 24 * 3600

  # Exceptions that don't mark the task as failed but as retry
  retry_on_exceptions = (
    pymongo.errors.AutoReconnect,
    pymongo.errors.OperationFailure,
    pymongo.errors.ConnectionFailure,
    RetryInterrupt
  )

  # Exceptions that will make the task as cancelled
  cancel_on_exceptions = (
    CancelInterrupt
  )

  def __init__(self, job_id, worker=None, queue=None, start=False, fetch=False):
    self.worker = get_current_worker()
    self.queue = queue
    self.datestarted = None

    self.collection = connections.mongodb_jobs.mrq_jobs
    self.id = ObjectId(job_id)

    self.data = None
    self.task = None
    self.greenlet_switches = 0
    self.greenlet_time = 0

    if start:
      self.fetch(start=True, full_data=False)
    elif fetch:
      self.fetch(start=False, full_data=False)

  def exists(self):
    return bool(self.collection.find_one({"_id": self.id}, fields={"_id": 1}))

  def fetch(self, start=False, full_data=True):
    """ Get the current job data and possibly flag it as started. """

    if full_data:
      fields = None
    else:
      fields = {
        "_id": 0,
        "path": 1,
        "params": 1,
        "status": 1
      }

    if start:
      self.datestarted = datetime.datetime.utcnow()
      self.data = self.collection.find_and_modify({
        "_id": self.id,
        "status": {"$nin": ["cancel"]}
      }, {"$set": {
        "status": "started",
        "datestarted": self.datestarted,
        "worker": self.worker.name
      }}, fields=fields)
    else:
      self.data = self.collection.find_one({
        "_id": self.id
      }, fields=fields)

    if self.data is None:
      log.info("Job %s not found in MongoDB or status was cancelled!" % self.id)
    else:
      task_def = get_current_config().get("tasks", {}).get(self.data["path"]) or {}
      self.timeout = task_def.get("timeout", self.timeout)
      self.result_ttl = task_def.get("result_ttl", self.result_ttl)

    return self

  def save_status(self, status, result=None, traceback=None, dateretry=None, queue=None, w=1):

    now = datetime.datetime.utcnow()
    updates = {
      "status": status,
      "dateupdated": now
    }

    if self.datestarted:
      updates["totaltime"] = (now - self.datestarted).total_seconds()
    if result is not None:
      updates["result"] = result
    if traceback is not None:
      updates["traceback"] = traceback
    if dateretry is not None:
      updates["dateretry"] = dateretry
    if queue is not None:
      updates["queue"] = queue
    if get_current_config().get("trace_greenlets"):
      current_greenlet = gevent.getcurrent()
      updates["time"] = current_greenlet._trace_time
      updates["switches"] = current_greenlet._trace_switches

    # Make the job document expire
    if status in ("success", "cancel"):
      updates["dateexpires"] = now + datetime.timedelta(seconds=self.result_ttl)

    self.collection.update({
      "_id": self.id
    }, {"$set": updates}, w=w)

    if self.data:
      self.data.update(updates)

  def save_retry(self, exc, traceback=None):

    countdown = 24 * 3600

    if isinstance(exc, RetryInterrupt) and exc.countdown is not None:
      countdown = exc.countdown

    queue = None
    if isinstance(exc, RetryInterrupt) and exc.queue:
      queue = exc.queue

    if countdown == 0:
      self.requeue(queue=queue)
    else:
      self.save_status(
        "retry",
        traceback=traceback,
        dateretry=datetime.datetime.utcnow() + datetime.timedelta(seconds=countdown),
        queue=queue
      )

  def retry(self, queue=None, countdown=None, max_retries=None):

    if self.task.cancel_on_retry:
      raise CancelInterrupt()
    else:
      exc = RetryInterrupt()
      exc.queue = queue
      exc.countdown = countdown
      raise exc

  def cancel(self):
    self.save_status("cancel")

  def requeue(self, queue=None):

    if not queue:
      if not self.data or not self.data.get("queue"):
        self.fetch(full_data=True)  # TODO only fetch queue?
      queue = self.data["queue"]

    self.save_status("queued", queue=queue)

    # Between these two lines, jobs can become "lost" too.

    Queue(queue).enqueue_job_ids([str(self.id)])

  def perform(self):
    """ Loads and starts the main task for this job, the saves the result. """

    if self.data is None:
      return

    log.debug("Starting %s(%s)" % (self.data["path"], self.data["params"]))
    task_class = load_class_by_path(self.data["path"])

    self.task = task_class()

    self.task.is_main_task = True

    result = self.task.run(self.data["params"])

    self.save_status("success", result)

    if get_current_config().get("trace_greenlets"):

      # TODO: this is not the exact greenlet_time measurement because it doesn't
      # take into account the last switch's time. This is why we force a last switch.
      # This does cause a performance overhead. Instead, we should print the
      # last timing directly from the trace() function in context?

      gevent.sleep(0)
      current_greenlet = gevent.getcurrent()
      log.debug("Job %s success: %0.6fs total, %0.6fs in greenlet, %s switches" % (
        self.id, (datetime.datetime.utcnow() - self.datestarted).total_seconds(),
        current_greenlet._trace_time, current_greenlet._trace_switches - 1)
      )
    else:
      log.debug("Job %s success: %0.6fs total" % (
        self.id, (datetime.datetime.utcnow() - self.datestarted).total_seconds()
      ))

  def wait(self, poll_interval=1, timeout=None, full_data=False):
    """ Wait for this job to finish. """

    collection = connections.mongodb_jobs.mrq_jobs

    end_time = None
    if timeout:
      end_time = time.time() + timeout

    while (end_time is None or time.time() < end_time):

      job_data = collection.find_one({
        "_id": ObjectId(self.id),
        "status": {"$nin": ["started", "queued"]}
      }, fields=({
        "_id": 0,
        "result": 1,
        "status": 1
      } if not full_data else None))

      if job_data:
        return job_data

      time.sleep(poll_interval)

    raise Exception("Waited for job result for %ss seconds, timeout." % timeout)



from mrq.queue import send_task, Queue
from mrq.task import Task
from mrq.job import Job
from mrq.context import log, connections
import datetime


class RequeueInterruptedJobs(Task):
  """ Requeue jobs that were marked as status=interrupt when a worker got a SIGTERM. """

  def run(self, params):
    return send_task("mrq.basetasks.utils.JobAction", {
      "status": "interrupt",
      "action": "requeue"
    }, sync=True)


class RequeueRetryJobs(Task):
  """ Requeue jobs that were marked as retry. """

  def run(self, params):
    return send_task("mrq.basetasks.utils.JobAction", {
      "status": "retry",
      "dateretry": {"$lte": datetime.datetime.utcnow()},
      "action": "requeue"
    }, sync=True)


class RequeueStartedJobs(Task):
  """ Requeue jobs that were marked as status=started and never finished.

      That may be because the worker got a SIGKILL or was terminated abruptly. The timeout parameter
      of this task is in addition to the task's own timeout.
  """

  def run(self, params):

    additional_timeout = params.get("timeout", 300)

    stats = {
      "requeued": 0,
      "started": 0
    }

    # There shouldn't be that much "started" jobs so we can quite safely iterate over them.
    self.collection = connections.mongodb_jobs.mrq_jobs
    for job_data in self.collection.find({"status": "started"}, fields={"_id": 1, "datestarted": 1}):
      job = Job(job_data["_id"])

      stats["started"] += 1

      expire_date = datetime.datetime.utcnow() - datetime.timedelta(seconds=job.timeout + additional_timeout)

      if job_data["datestarted"] < expire_date:
        log.debug("Requeueing job %s" % job.id)
        job.requeue()
        stats["requeued"] += 1

    return stats


class RequeueLostJobs(Task):
  """ Requeue jobs that were queued but don't appear in Redis anymore.

      They could have been lost either by a Redis flush or by a worker interrupt between
      redis.blpop and mongodb.update
  """

  def run(self, params):

    self.collection = connections.mongodb_jobs.mrq_jobs

    # If there are more than this much items on the queue, we don't try to check if our mongodb
    # jobs are still queued.
    max_queue_items = params.get("max_queue_items", 1000)

    stats = {
      "fetched": 0,
      "requeued": 0
    }

    for job_data in self.collection.find({
      "status": "queued"
    }, fields={"_id": 1, "queue": 1}).sort([("_id", 1)]).batch_size(100):

      stats["fetched"] += 1

      queue = Queue(job_data["queue"])
      queue_size = queue.size()
      if queue_size > max_queue_items:
        log.info("Stopping because queue %s has %s items" % (queue, queue_size))
        break

      queue_jobs_ids = set(queue.list_job_ids(limit=max_queue_items + 1))
      if len(queue_jobs_ids) >= max_queue_items:
        log.info("Stopping because queue %s actually had more than %s items" % (queue, len(queue_jobs_ids)))
        break

      if str(job_data["_id"]) in queue_jobs_ids:
        log.info("Stopping because we found job %s in redis" % job_data["_id"])
        break

      # At this point, this job is not on the queue and we're sure the queue is less than max_queue_items
      # We can safely requeue the job.
      log.info("Requeueing %s on %s" % (job_data["_id"], queue.id))

      stats["requeued"] += 1
      job = Job(job_data["_id"])
      job.requeue(queue=job_data["queue"])

    return stats


# class CleanupStoppedWorkers(Task):
#   """ Remove stopped or timeouted workers from mongodb.mrq_workers. """

#   def run(self, params):




from .logger import LoggerInterface
import gevent
import urlparse
import redis as pyredis
import re
import time
from .utils import LazyObject

try:
  # MongoKit's Connection object is just a wrapped MongoClient.
  from mongokit import Connection as MongoClient
except:
  from pymongo import MongoClient


# greenletid => Job object
_GREENLET_JOBS_REGISTRY = {}

_WORKER = None

_CONFIG = None

# Global log object, usable from all tasks
log = LoggerInterface(None, job="current")


def set_current_job(job):
  current = gevent.getcurrent()
  _GREENLET_JOBS_REGISTRY[id(current)] = job

  current.__dict__["_trace_time"] = 0
  current.__dict__["_trace_switches"] = 0


def get_current_job(greenlet_id=None):
  if greenlet_id is None:
    greenlet_id = id(gevent.getcurrent())
  return _GREENLET_JOBS_REGISTRY.get(greenlet_id)


def set_current_worker(worker):
  global _WORKER
  _WORKER = worker
  set_current_config(worker.config)


def get_current_worker():
  global _WORKER
  return _WORKER


def set_current_config(config):
  global _CONFIG
  _CONFIG = config
  log.quiet = config["quiet"]


def get_current_config():
  global _CONFIG
  return _CONFIG


def retry_current_job(**kwargs):
  get_current_job().retry(**kwargs)


def _connections_factory(attr):

  config = get_current_config()

  # Connection strings may be stored directly in config
  config_obj = config.get(attr)

  if attr == "redis":

    if type(config_obj) in [str, unicode]:

      urlparse.uses_netloc.append('redis')
      redis_url = urlparse.urlparse(config_obj)

      log.info("Connecting to Redis at %s..." % redis_url.hostname)

      redis_pool = pyredis.ConnectionPool(
        host=redis_url.hostname,
        port=redis_url.port,
        db=0,
        password=redis_url.password
      )
      return pyredis.StrictRedis(connection_pool=redis_pool)

    # Let's just assume we got a StrictRedis-like object!
    else:
      return config_obj

  elif attr.startswith("mongodb"):

    if type(config_obj) in [str, unicode]:

      (mongoAuth, mongoUsername, mongoPassword, mongoHosts, mongoDbName) = re.match(
        "mongodb://((\w+):(\w+)@)?([\w\.:,-]+)/([\w-]+)", config_obj).groups()

      log.debug("Connecting to MongoDB at %s/%s..." % (mongoHosts, mongoDbName))

      value = MongoClient(mongoHosts)[mongoDbName]
      if mongoUsername:
        value.authenticate(mongoUsername, mongoPassword)

      return value

    # Let's just assume we got a MongoDB-like object!
    else:
      return config_obj

connections = LazyObject()
connections.add_factory(_connections_factory)
del _connections_factory


def enable_greenlet_tracing():

  # Tracing seems to cause a 2-5% performance loss.

  import greenlet
  greenlet.GREENLET_USE_TRACING = True

  def trace(*args):

    time_since_last_switch = time.time() - trace.last_switch

    # Record the time of the current switch
    trace.last_switch = time.time()

    if args[0] == "switch":
      # We are switching from the greenlet args[1][0] to the greenlet args[1][1]
      args[1][0].__dict__.setdefault("_trace_time", 0)
      args[1][0].__dict__["_trace_time"] += time_since_last_switch
      args[1][0].__dict__.setdefault("_trace_switches", 0)
      args[1][0].__dict__["_trace_switches"] += 1

    elif args[0] == "throw":
      pass

  trace.last_switch = time.time()

  greenlet.settrace(trace)



from mrq.task import Task
from mrq.context import log, get_current_job, get_current_worker, get_current_config


class GetContext(Task):
  def run(self, params):
    log.info("Getting context info...")
    return {
      "job_id": get_current_job().id,
      "worker_id": get_current_worker().id,
      "config": get_current_config()
    }




class Task(object):

  cancel_on_retry = False
  cancel_on_exception = False
  cancel_on_timeout = False

  # Are we the first task that a Job called?
  is_main_task = False

  def __init__(self):
    pass

  def run(self, params):
    raise NotImplementedError



from setuptools import setup, find_packages
import os
from pip.req import parse_requirements


def get_requirements():
    # found at http://stackoverflow.com/questions/14399534/how-can-i-reference-requirements-txt-for-the-install-requires-kwarg-in-setuptool
    # reqs is a list of requirement
    # e.g. ['django==1.5.1', 'mezzanine==1.4.6']
    reqs = []
    for filename in ["requirements.txt", "requirements-dashboard.txt"]:
        # parse_requirements() returns generator of pip.req.InstallRequirement objects
        install_reqs = parse_requirements(filename)
        reqs += [str(ir.req) for ir in install_reqs]
    return reqs


def get_version():
    basedir = os.path.dirname(__file__)
    with open(os.path.join(basedir, 'mrq/version.py')) as f:
        locals = {}
        exec(f.read(), locals)
        return locals['VERSION']
    raise RuntimeError('No version info found.')

setup(
    name="mrq",
    include_package_data=True,
    packages=find_packages(exclude=['tests']),
    version=get_version(),
    description="Mongo Redis Queue",
    author="Pricing Assistant",
    license='MIT',
    author_email="contact@pricingassistant.com",
    url="http://github.com/pricingassistant/mrq",
    # download_url="http://chardet.feedparser.org/download/python3-chardet-1.0.1.tgz",
    keywords=["worker", "task", "distributed", "queue", "asynchronous"],
    platforms='any',
    entry_points={
        'console_scripts': [
            'mrq-worker = mrq.bin.mrq_worker:main',
            'mrq-run = mrq.bin.mrq_run:main',
            'mrq-dashboard = mrq.dashboard.app:main'
        ]
    },
    zip_safe=False,
    install_requires=get_requirements(),
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        #'Development Status :: 1 - Planning',
        #'Development Status :: 2 - Pre-Alpha',
        'Development Status :: 3 - Alpha',
        #'Development Status :: 4 - Beta',
        #'Development Status :: 5 - Production/Stable',
        #'Development Status :: 6 - Mature',
        #'Development Status :: 7 - Inactive',
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Utilities"
    ],
    long_description=open("README.md").read()
)



from time import sleep
from mrq.task import Task
from mrq.context import log, retry_current_job, connections
import urllib2


class Add(Task):

  def run(self, params):

    log.info("adding", params)
    res = params.get("a", 0) + params.get("b", 0)

    if params.get("sleep", 0):
      log.info("sleeping", params.get("sleep", 0))
      sleep(params.get("sleep", 0))

    return res


class TimeoutFromConfig(Add):
  pass


class TimeoutFromConfigAndCancel(Add):
  cancel_on_timeout = True


class Fetch(Task):
  def run(self, params):

    f = urllib2.urlopen(params.get("url"))
    t = f.read()
    f.close()

    return len(t)


class Retry(Task):

  def run(self, params):

    log.info("Retrying in %s on %s" % (params.get("countdown"), params.get("queue")))

    connections.mongodb_logs.tests_inserts.insert(params)

    if params.get("cancel_on_retry"):
      self.cancel_on_retry = params.get("cancel_on_retry")

    retry_current_job(queue=params.get("queue"), countdown=params.get("countdown"))

    raise Exception("Should not be reached")


class RaiseException(Task):

  def run(self, params):

    sleep(params.get("sleep", 0))

    raise Exception(params.get("message", ""))


class ReturnParams(Task):

  def run(self, params):

    sleep(params.get("sleep", 0))

    return params


class MongoInsert(Task):

  def run(self, params):

    connections.mongodb_logs.tests_inserts.insert(params)


MongoInsert2 = MongoInsert





def add_standard_worker_arguments(parser):
  parser.add_argument('--maxtasks', default=0, type=int, action='store', help='Gevent: max number of tasks to do before quitting. Temp workaround for memory leaks')
  parser.add_argument('--mongodebug', action='store_true', default=False, help='Print all Mongo requests')
  parser.add_argument('--objgraph', action='store_true', default=False, help='Start objgraph to debug memory after each task')
  parser.add_argument('--profile', action='store_true', default=False, help='Run profiling')
  parser.add_argument('--name', default=None, help='Specify a different name')
  parser.add_argument('queues', nargs='*', help='The queues to listen on (default: \'default\')')



import re
import importlib
import time


def group_iter(iterator, n=2):
  """ Given an iterator, it returns sub-lists made of n items.
  (except the last that can have len < n)

  """
  accumulator = []
  for item in iterator:
    accumulator.append(item)
    if len(accumulator) == n:
      yield accumulator
      accumulator = []

  # Yield what's left
  if len(accumulator) != 0:
    yield accumulator


# http://code.activestate.com/recipes/578231-probably-the-fastest-memoization-decorator-in-the-/
def memoize(f):
  """ Memoization decorator for a function taking one or more arguments. """
  class memodict(dict):
    def __getitem__(self, *key):
      return dict.__getitem__(self, key)

    def __missing__(self, key):
      ret = self[key] = f(*key)
      return ret

  return memodict().__getitem__


def memoize_single_argument(f):
  """ Memoization decorator for a function taking a single argument """
  class memodict(dict):
    def __missing__(self, key):
      ret = self[key] = f(key)
      return ret
  return memodict().__getitem__


@memoize_single_argument
def load_class_by_path(taskpath):
  """ Given a taskpath, returns the main task class. """

  return getattr(importlib.import_module(re.sub(r"\.[^.]+$", "", taskpath)), re.sub(r"^.*\.", "", taskpath))


def lazyproperty(fn):
  attr_name = '_lazy_' + fn.__name__

  @property
  def _lazyprop(self):
    if not hasattr(self, attr_name):
      setattr(self, attr_name, fn(self))
    return getattr(self, attr_name)
  return _lazyprop


# http://code.activestate.com/recipes/576655-wait-for-network-service-to-appear/
def wait_for_net_service(server, port, timeout=None):
  """ Wait for network service to appear
      @param timeout: in seconds, if None or 0 wait forever
      @return: True of False, if timeout is None may return only True or
               throw unhandled network exception
  """
  import socket
  import errno

  s = socket.socket()
  if timeout:
    from time import time as now
    # time module is needed to calc timeout shared between two exceptions
    end = now() + timeout

  while True:
    try:
      if timeout:
        next_timeout = end - now()
        if next_timeout < 0:
          return False
        else:
          s.settimeout(next_timeout)

      s.connect((server, port))

    except socket.timeout, err:
      # this exception occurs only if timeout is set
      if timeout:
        return False

    except Exception, err:
      # catch timeout exception from underlying network library
      # this one is different from socket.timeout
      if type(err.args) != tuple or err[0] != errno.ETIMEDOUT:
        pass  # raise
    else:
      s.close()
      return True
    time.sleep(0.1)


class LazyObject(object):
  """ Lazy-connection class. Connections will only be initialized when first used. """

  def __init__(self):
    self._factories = []
    self._attributes_via_factories = []

  def add_factory(self, factory):
    self._factories.append(factory)

  # This will be called only once, when the attribute is still missing
  def __getattr__(self, attr):

    for factory in self._factories:
      value = factory(attr)
      if value is not None:
        self._attributes_via_factories.append(attr)
        self.__dict__[attr] = value
        return value

  def reset(self):
    # TODO proper connection close?
    for attr in self._attributes_via_factories:
      del self.__dict__[attr]
    self._attributes_via_factories = []
